-- QBasic I/O Routines for eBasic
-- Printing and Input
--      (c) 1997 David Cuny


without warning
without trace
without type_check

include msgbox.e
include graphics.e
include wildcard.e
include get.e
include int24.e     -- Jacques Deschenes' DOS error trapping

-- i/o checking
procedure check_io()

    -- check for DOS error
    integer err

    err = CheckIo()
    if err != 0 then
        abort_error( "IO Error", IO_ERRORS[ err ] )
    end if

end procedure


-- Tab
constant TAB_SIZE = 14

-- File handle data structure
--
-- In order for EOF to work, a file read has to happen.
-- The data read by that EOF is stored in F_BUFFER.
--
--  File -1     = PRN
--  File 0      = SCREEN
--

global constant
    MAX_FILES   = 16,   -- maximum number of files
    F_HANDLE    = 1,    -- handle
    F_MODE      = 2,    -- mode
    F_EOF       = 3,    -- EOF flag
    F_DIRTY     = 4,    -- true if buffer contains data
    F_BUFFER    = 5,    -- value of last read buffer
    F_TAB       = 6     -- position of cursor


-- set up file handles
    global sequence file
    file = repeat( { 0, 0, "", 0, "", 1 }, MAX_FILES+2 )

-- for multiple values returned
    global sequence inputBuffer 

--
-- FILE SUPPORT
--


-----------------------------------------------------------------------------
global function file_get_pos( integer device )
    -- return column cursor is on
    
    sequence s

    -- screen?   
    if device = - 1 then
        -- printer
        return file[MAX_FILES+1][F_TAB]
    
    elsif device = 0 then
        -- screen
        return file[MAX_FILES+2][F_TAB]

    else
        -- file
        return file[device][F_TAB]
        
    end if

end function


-----------------------------------------------------------------------------
global procedure file_set_pos( integer device, integer pos )
    -- return column cursor is on
    
    sequence s

    -- screen?   
    if device = - 1 then
        -- printer
        file[MAX_FILES+1][F_TAB] = pos
    
    elsif device = 0 then
        -- screen
        file[MAX_FILES+2][F_TAB] = pos

    else
        -- file
        file[device][F_TAB] = pos
        
    end if

end procedure


-----------------------------------------------------------------------------
global procedure file_open( sequence name, sequence mode, integer device )

    -- open the given file; set up buffers & flags

    -- screen
    if device = 0 then
        -- already open
        return
    end if

    -- printer
    if device = -1 then
        -- printer    
        name = "PRN"
        device = MAX_FILES + 1
    end if
    
    file[device][F_HANDLE]      = open( name, mode )
    file[device][F_MODE]        = mode
    file[device][F_EOF]         = 0
    file[device][F_DIRTY]       = 0
    file[device][F_BUFFER]      = ""
    file[device][F_TAB]         = 1
    
    if file[device][F_HANDLE] = -1 then
        abort_error( "File Error", sprintf( "File \"%s\" not found.", {name} ) )
    end if

end procedure

-----------------------------------------------------------------------------
global function file_read( integer device )

    object result, char

    -- read from the given file

    -- invalid device?
    if file[device][F_HANDLE] < 1 then

        -- error
        abort_error( "File Error", "Bad file name or number" )
        
     -- something in buffer
    elsif file[device][F_DIRTY] then

            -- reset flag
            file[device][F_DIRTY] = 0

            -- return buffered data
            return file[device][F_BUFFER]

     -- end of file
    elsif file[device][F_EOF] then
    
        -- return eof marker
        return {-1}


    -- read the file
    elsif compare( file[device][F_MODE], "r" ) = 0 then

        -- read with gets
        -- faster, but less compatible read

        result = gets( file[device][F_HANDLE] )
        check_io()

        if compare( result, -1 ) = 0 then
            -- end of file
            return {-1}

        else
            -- good read
            return result

        end if


    elsif compare( file[device][F_MODE], "rb" ) = 0 then

        -- read until new line or eof
        -- slower, but more compatible

        char = {}    
        result = ""
        while 1 do

            char = getc( file[device][F_HANDLE] )
            check_io()

            -- append, unless cr
            if char != 10 then
                result = result & char
            end if

            -- exit loop?
            if char = -1 then
                -- end of file
                exit

            elsif char = 13 then
                -- line feed
                exit

            end if

        end while

        -- return result
        return result

    else

        -- error
        abort_error( "File Error", "File not open for read mode." )

    end if
    
end function


-----------------------------------------------------------------------------
global function file_eof( integer device )

    object result

    -- return true if eof
    -- since the file needs to be read to see if it's the eof,
    -- the data is buffered, and read by file_read().
    
    if file[device][F_HANDLE] < 1 then
        -- error
        abort_error( "File Error", "Bad file name or number." )
        
    elsif file[device][F_DIRTY] then
        -- data pending in buffer; not at end
        return 0

    elsif file[device][F_EOF] then
        -- already encountered
        return -1                 

    else
        -- read from file
        result = file_read( device )

        -- eof only?
        if compare( result, {-1} ) = 0 then

            -- true
            return -1

        else
            -- not end of file

            -- buffer results
            file[device][F_DIRTY] = -1
            file[device][F_BUFFER] = result

            -- return false
            return 0
                
        end if

    end if
    
end function    

-----------------------------------------------------------------------------
global procedure file_close( integer device )
    -- close all the files
    
 
    -- close file
    close( file[device][F_HANDLE] )
    
    -- clear the flag
    file[device][F_HANDLE] = 0
    
end procedure
    
-----------------------------------------------------------------------------
global procedure file_close_all()

    -- close all open files, clear flags
    
    for i = 1 to MAX_FILES do
        -- open?
        if file[i][F_HANDLE] > 0 then

            -- close it
            file_close( i )

        end if
    end for
    
end procedure


-----------------------------------------------------------------------------
--
-- PRINTING
--



-----------------------------------------------------------------------------
global procedure print_str( integer device, sequence s )

    -- this is the primary print routine
    -- make sure it fits
    if device = 0 then

        -- special case: hardware has forced a wrap on the edge
        if compare( s, "\n" ) = 0 then
            -- cr
            if file_get_pos( device ) >= 80 then
                -- has already happened
                file_set_pos( 0, 1 )
                return
            end if

        else
            -- will it fit on the screen?
            if file_get_pos( device ) + length( s ) >= 80 then
                -- print a new line
                puts( 1, "\n" )
                file_set_pos( 0, 1 )
            end if

        end if
    
    elsif device > 0 then

        if file[device][F_HANDLE] <= 0 then
            abort_error( "File Error", "Attempt to write to unopened file." )            

        elsif not find( 'w', file[device][F_MODE] ) then
            abort_error( "File Error", "Attempt to write to file in read mode." )

        end if
    end if

    -- print to the device 
    if      device = -1 then    puts( "PRN", s )
    elsif   device = 0  then    puts( 1, s )
    else    puts( file[device][F_HANDLE], s )    
    end if

    -- new position
    file_set_pos( device, file_get_pos( device ) + length( s ) )    

end procedure


-----------------------------------------------------------------------------
global procedure print_num( integer device, atom n )
    -- numeric print

    if n < 0 then    
        -- no leading space on negatives
        print_str( device, sprintf( "%g ", {n} ) )        
    else
        -- 
        print_str( device, sprintf( " %g ", {n} ) )
    end if

end procedure


-----------------------------------------------------------------------------
global procedure print_cr( integer device )
    -- print tab        
    
    -- print
    print_str( device, "\n" )
    file_set_pos( device, 1 )
    
end procedure


-----------------------------------------------------------------------------
global procedure print_tab( integer device )
    -- print tab
    integer n, tabPos

    -- get current position
    tabPos = file_get_pos( device )

    -- calculate new position    
    n = TAB_SIZE - remainder( tabPos-1, TAB_SIZE )

    -- printing to screen?
    if device = 0 then
        -- off the edge?
        if tabPos + n > 80 then
            -- line feed
            print_cr( device )
            -- no spaces
            n = 0
        end if
    end if

    -- print spaces
    print_str( device, repeat( ' ', n ) )
    
end procedure

-----------------------------------------------------------------------------
global procedure print_tab_to( integer device, integer tab )
    -- tab to a given column
    
    integer n, tabPos

    -- get current tab position
    tabPos = file_get_pos( device )

    -- past position?    
    if tab < tabPos then

        -- do a cr
        print_cr( device )
        
        -- print spaces
        print_str( device, repeat( ' ', tab-1 ) )
        
    else

        -- calculate number of spaces
        n = tab - tabPos
        
        -- print spaces
        print_str( device, repeat( ' ', n ) )

    end if

end procedure


-----------------------------------------------------------------------------
function output_mask( sequence mask, integer size, sequence data, integer i )

    -- returns formatted element data[i]
    -- if no data element, just print blanks

    -- ensure that there is data
    if i > length( data ) then
        -- just return spaces
        return repeat( ' ', size )
    else
        -- use mask
        return sprintf( mask, {data[i]} ) 
    end if
    
    
end function
        

-----------------------------------------------------------------------------
global procedure qb_print_using( integer device, sequence s, sequence data )

    -- create output based on text mask
    

    integer numMask, strMask, special, digits, size, dataAt
    sequence mask, out, typeChar

    numMask = 0     -- flag if in a numeric mask
    strMask = 0     -- flag if in a string mask      
    typeChar = ""   -- holds numeric data type
    special = 0     -- flag if next character is literal
    digits = 0      -- number of digits encountered in mask
    size = 0        -- size of mask; used if data is missing
    dataAt = 0      -- which element in data sequence is current
    mask = ""       -- mask for sprintf
    out = ""        -- resulting output

    -- pad end of text
    s = s & " "

    for i = 1 to length( s ) do
        if special then
            -- add "special" char to output
            out = out & s[i]
    
            -- kill flag
            special = 0

        elsif s[i] = '_' then
            -- "special" character flag
            special = -1
    
            -- end of mask?
            if numMask
            or strMask then

                -- output data
                dataAt = dataAt + 1            
                out = out & output_mask( mask, size, data, dataAt )
                
                -- clear flags
                numMask = 0
                strMask = 0
            
            end if
    
        elsif s[i] = '-' 
        and   numMask then            
            -- trailing minus sign            
            
            -- check sign
            dataAt = dataAt + 1
            
            -- data exists?
            if dataAt <= length( data ) then

                -- add trailing minus/space
                if data[dataAt] < 0 then         
                    -- drop sign, place in mask
                    data[dataAt] = - data[dataAt]
                    mask = mask & typeChar & "-"
                else
                    -- trailing space instead of sign
                    mask = mask & typeChar & " "
                end if                

            end if            

            -- add 1 byte to size
            size = size + 1

            -- build it (already moved dataAt)
            out = out & output_mask( mask, size, data, dataAt )                                      

            -- clear flag
            numMask = 0

        elsif s[i] = '+' then
            -- leading sign

            -- pending numeric mask?
            if numMask then
                -- finish mask
                mask = mask & sprintf( "%d%s", {digits,typeChar} )

                -- add to output                
                dataAt = dataAt + 1
                out = out & output_mask( mask, size, data, dataAt )

            end if

            -- flag mask
            numMask = -1

            -- start mask
            mask = "%+"

            -- clear size
            digits = 0
            size = 1
                
            -- assume decimal
            typeChar = "d"


        elsif s[i] = '#' then
    
            -- start of mask?
            if not numMask then
            
                -- start mask
                mask = "%"

                -- clear size
                size = 1
                
                -- assume decimal
                typeChar = "d"
                
            end if
    
            -- increment digits, size            
            digits = digits + 1
            size = size + 1
    
            -- set mask flag
            numMask = -1
    
        elsif s[i] = '.' then
            -- has mask been started?
            if not numMask then
            
                -- start mask
                mask = "%0."
                
                -- clear size
                size = 0

            else    
                
                -- build integer portion of mask
                mask = mask & sprintf( "%d.", {digits} )
                
            end if
            
            -- set mask flag
            numMask = -1
    
            -- clear digit count
            digits = 0
            
            -- increment size
            size = size + 1
            
            -- floating point
            typeChar = "f"

        elsif s[i] = ','
        and   numMask then
            -- Euphoria does not support commas
            -- but keep the width the same
            digits = digits + 1
            

        elsif s[i] = ' '
        and   strMask then
            -- add to size
            size = size + 1

        elsif s[i] = '\\' then
            -- pending numeric mask?
            if numMask then
                -- finish mask
                mask = mask & sprintf( "%d%s", {digits,typeChar} )

                -- add to output                
                dataAt = dataAt + 1
                out = out & output_mask( mask, size, data, dataAt )

                -- clear flag                
                numMask = 0                
                
            end if
        
            -- start, or end of string mask
            if strMask then              
                -- end of string mask
                
                -- build mask                
                mask = sprintf( "%%.%ds", {size} )
                
                -- use mask
                dataAt = dataAt + 1         
                out = out & output_mask( mask, size, data, dataAt )

                -- clear flag
                strMask = 0
    
            else
                -- start of string mask
                strMask = -1
    
                -- add to count, include end char
                size = 2
    
            end if
        else
            -- end of numeric mask?

            if numMask then
                -- finish mask
                mask = mask & sprintf( "%d%s", {digits,typeChar} )

                -- add to output                
                dataAt = dataAt + 1
                out = out & output_mask( mask, size, data, dataAt )

                -- clear flag                
                numMask = 0                
                
            end if
    
            -- alpha?
            if s[i] = '&' then
                -- sequence 
                dataAt = dataAt + 1                
                out = out & output_mask( "%s", size, data, dataAt )
    
            elsif s[i] = '!' then
                -- 1 char
                dataAt = dataAt + 1
                out = out & output_mask( "%.1s", size, data, dataAt )
    
            else
                -- normal character
                out = out & s[i]
            end if
        end if
            
    end for

    -- remove extra space at end
    out = out[1..length(out)-1]

    -- write to device    
    print_str( device, out )

end procedure



--
-- INPUT SUPPORT
--

-----------------------------------------------------------------------------
function input_gets()

    -- Accept input from screen. Handles screen wrapping, scrolling

    integer key, start, row, col, long
    sequence startPos, s
    
    -- clear string
    s = ""

    -- get initial screen position
    startPos = get_position()
    start = ((startPos[1]-1)*80) + startPos[2] - 2

    while 1 do
        -- get a key
        key = wait_key()

        if key = 13 then
            -- enter
            exit

        elsif key = 27 then
            s = ""
            exit

        elsif key = 8
        or    key = 331 then
            -- backspace or left arrow

            if length(s) then

                -- calculate screen position            
                long = start + length(s)
                row = floor( long/80 )+1
                col = remainder( long, 80 )+1
                
                -- erase char at current position
                position( row, col )
                puts( 1, " " )

                -- position cursor
                position( row, col )

                -- delete a char
                s = s[1..length(s)-1]

            end if

        elsif key >= ' '            -- higher than a space
        and   key <= 255            -- real ascii char
        and length( s ) < 256 then  -- QBasic limits to 255 chars
            
            -- append
            s = s & key

            -- calculate screen position
            long = start + length(s)
            row = floor( long/80 )+1
            col = remainder( long, 80 )+1

            -- reached bottom of screen?
            if row = 25 and col = 80 then
                -- line feed will happen automatically
                -- adjust positions
                startPos[1] = startPos[1]-1
                start = start - 80
            end if

            -- show key                
            position( row, col )
            puts( 1, key )

        end if

    end while

    return s

end function


-----------------------------------------------------------------------------
function input_split( sequence s )
    -- split input into parts
    integer quote
    sequence words, word

    -- setup
    words = {}      -- accumulated words
    word = ""       -- accumulated word
    quote = 0       -- quote mark

    -- scan
    for i = 1 to length( s ) do
        if quote and s[i] != '"' then
            word = word & s[i]

        elsif s[i] = '"' then
            -- toggle flag
            quote = not quote

            -- append to word
            word = word & s[i]

        elsif s[i] = ',' then
            words = append( words, word )
            word = ""

        else
            word = word & s[i]

        end if
            
    end for

    -- final word
    words = append( words, word )

    return words

end function


-----------------------------------------------------------------------------
global function input_screen( sequence prompt, integer qFlag, integer crFlag,
                                integer lineFlag, sequence mask )
                                
    -- get input from screen
    --  if length of mask = 1 or lineFlag = true,
    --      returns a single element
    --  else
    --      returns a sequence of elements
    --
    -- Parameter    Function
    -- ----------   ----------------------------------------
    -- prompt       Prompt to display
    -- qFlag        Add "? " to prompt
    -- crFlag       Add line feed after user presses return
    -- lineFlag     Line input, accept as entered
    -- mask         string containing expected type
    --                  n = numeric
    --                  s = string
    

    integer flag
    object s, result

    -- loop until good input
    while 1 do

        -- prompt
        puts( 1, prompt )
        
        -- optional question mark
        if qFlag then
            puts( 1, "? " )
        end if   

        -- get input
        s = input_gets()

        -- optional line feed
        if crFlag then
            -- line feed
            puts( 1, "\n" )
        end if

        -- accept if line input
        if lineFlag then
            return s
        end if

        -- split
        s = input_split( s )
        -- correct count?
        if length( s  ) = length( mask ) then
        
            -- flag initially true
            flag = 1
            
            -- validate
            for i = 1 to length( s ) do
                -- need to validate numeric
                if mask[i] = 'n' then
                
                    -- attempt to convert
                    result = value( s[i] )
                    -- number?
                    if result[1] = GET_SUCCESS then
                        -- save number
                        s[i] = result[2]
                    elsif length( s[i] ) = 0 then
                        -- blank is zero
                        s[i] = 0
                    else
                        -- flag as failed
                        flag = 0
                    end if
                    
                end if
            end for 
            
            -- all elements good?
            if flag then

                -- return validated string
                if length( mask ) = 1 then
                    -- return single element
                    return s[1]
                else
                    -- return list
                    return s
                end if

            end if
        
        end if
        
        -- error
        puts( 1, "\nRedo from start\n" )        

    end while

end function


-----------------------------------------------------------------------------
global function input_file( integer device, integer lineFlag, sequence mask )
    -- get input from file

    sequence s1, s2
    object result

    -- build the default response
    s2 = repeat( "", length( mask ) )

    -- read from file
    s1 = file_read( device )

    -- remove last char (-1 or cr)
    s1 = s1[1..length(s1)-1]
    
    -- accept if line input
    if lineFlag then
        return s1
    end if

    -- split it
    s1 = input_split( s1 )
    
    -- validate
    for i = 1 to length( mask ) do
        -- past end of input?
        if i > length( s1 ) then
            -- create a number, if not a string
            if mask[i] = 'n' then
                s2[i] = 0
            end if
        else        
            -- number?
            if mask[i] = 'n' then
                -- attempt to convert
                result = value( s1[i] )
                if result[1] = GET_SUCCESS then
                    -- store number
                    s2[i] = result[2]
                else
                    -- default to 0
                    s2[i] = 0
                end if
            end if
        end if
    end for 

    -- return validated string
    if length( mask ) = 1 then
        -- return single element
        return s2[1]
    else
        -- return list
        return s2
    end if
    
end function


